package scanner;

%%

%{
  public static void main(String argv[]) throws java.io.IOException {
    Scanner scanner = new Scanner(System.in);
    while (scanner.yylex()) ;
  }
  
  protected void comment(String text)  { System.out.println("\t#-comment: "+text); }
  protected void integer(String text)  { System.out.println("\tinteger: "+text); }
  protected void floating(String text) { System.out.println("\tfloating point: "+text); }
  protected void keyword(String text)  { System.out.println("\tkeyword: "+text); }
  protected void name(String text)     { System.out.println("\tidentifier: "+text); }
  protected void character(char ch)    { System.out.println("\tcharacter: "+ch); }
%}

%public
%class Scanner
%type boolean
%eofval{
  return false;
%eofval}

alpha	=	[a-zA-Z_]
alnum	=	[a-zA-Z_0-9]
dec	=	[0-9]
sign	=	[+-]?
exp	=	([eE]{sign}{dec}+)
fp	=	({dec}+\.{dec}*{exp}?|"."{dec}+{exp}?|{dec}+{exp})
whitespace	=	[ \t\r\n\b\015]+
%%

#.*	{ comment(yytext().substring(1)); return true; }
{dec}+	{ integer(yytext()); return true; }
{fp}	{ floating(yytext()); return true; }
while	{ keyword(yytext()); return true; }
{alpha}{alnum}*	{ name(yytext()); return true; }
{whitespace}	{ }
.	{ character(yytext().charAt(0)); return true; }
