package states;
import java.io.IOException;

%%

%public
%class Scanner

%line

%type boolean
%function advance
%eofval{
  return false;
%eofval}

%{
  /** contains string during assembly.
    */
  protected StringBuffer body;

  /** current input.
    */
  protected int token;

  /** determines current input.
      @return token value or character.
    */
  public int token () {
    return token;
  }

  /** @return input line position, counted from 1.
    */
  public int yyline () {
    return yyline+1;
  }

  /** reports lexical error, can be overwritten in subclass.
    */
  protected void lexerror (String message) {
    System.err.println(yyline()+": "+message);
  }

  /** identifier screener, to be overwritten in subclass.
      @param name identifier.
      @return token value.
    */
  protected int name (String name) {
    System.out.println("name\t"+name);
    return -1;
  }
  
  /** integer screener, to be overwritten in subclass.
      @param integer value.
      @return token value.
    */
  protected int integer (int integer) {
    System.out.println("integer\t"+integer);
    return -1;
  }
  
  /** float screener, to be overwritten in subclass.
      @param floating value.
      @return token value.
    */
  protected int floating (float floating) {
    System.out.println("floating\t"+floating);
    return -1;
  }

  /** string screener, to be overwritten in subclass.
      @param body internal string value.
      @return token value.
    */
  protected int string (String body) {
    System.out.println("string\t"+body);
    return -1;
  }

  /** relational operator, to be overwritten in subclass.
      @return token value.
    */
  protected int le () {
    System.out.println("le");
    return -1;
  }
  
  /** relational operator, to be overwritten in subclass.
      @return token value.
    */
  protected int ge () {
    System.out.println("ge");
    return -1;
  }
  
  /** relational operator, to be overwritten in subclass.
      @return token value.
    */
  protected int ne () {
    System.out.println("ne");
    return -1;
  }

  /** test program.
    */
  public static void main (String args []) {
    Scanner scanner = new Scanner(System.in);
    try {
      while (scanner.advance())
        if (scanner.token >= 0)
          System.out.println("char:\t"+(char)scanner.token);
    } catch (IOException e) { System.err.println(e); System.exit(1); }
  }
%}

comment	= ("//".*)
space	= [\ \t\n\b\015]+
letter	= [A-Za-z_]
digit	= [0-9]
integer	= {digit}+
exp	= ([eE](\+|\-)?{digit}+)
floating	= ({digit}+\.{digit}*{exp}?|{digit}*\.{digit}+{exp}?|{digit}+{exp})
name	= ({letter}({letter}|{digit})*)

%state	STRING
%%

<YYINITIAL>{space}          { }
<YYINITIAL>{comment}        { }
<YYINITIAL>{integer}        { token = integer(Integer.valueOf(yytext()).intValue());
                              return true;
                            }
<YYINITIAL>{floating}       { token = floating(Float.valueOf(yytext()).floatValue());
                              return true;
                            }
<YYINITIAL>{name}           { token = name(yytext());
                              return true;
                            }
<YYINITIAL>\"               { yybegin(STRING);
                              body = new StringBuffer();
                            }
<YYINITIAL>"<="             { token = le(); return true; }
<YYINITIAL>">="             { token = ge(); return true; }
<YYINITIAL>"<>"             { token = ne(); return true; }
<YYINITIAL>.                { token = yytext().charAt(0); return true; }

<STRING>\"                  { yybegin(YYINITIAL);
                              token = string(body.toString());
                              return true;
                            }
<STRING>\\([0-7][0-7]?|[0-3][0-7][0-7]) {
                              body.append((char) Integer.parseInt(yytext().substring(1), 8));
                            }
<STRING>\\b                 { body.append('\b'); }
<STRING>\\f                 { body.append('\f'); }
<STRING>\\n                 { body.append('\n'); }
<STRING>\\r                 { body.append('\r'); }
<STRING>\\t                 { body.append('\t'); }
<STRING>\\.                 { body.append(yytext().charAt(1)); }
<STRING>\\\n                { }
<STRING>.                   { body.append(yytext().charAt(0)); }
<STRING>\n                  { lexerror("\\n in string constant"); }
